#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cptbl.h"
#include "charConv.h"

#define IS_SJIS_ZENKAKU(x) ( (0x80 < x && x < 0xa0) || (0xdf < x && x < 0xf0) )


//==============================================================
// VtgJIS-UCSϊe[ũ\[g
//--------------------------------------------------------------

static int mapSort(const void *a, const void *b)
{
	unsigned short	*ch1, *ch2;

	ch1 = (unsigned short*)a;
	ch2 = (unsigned short*)b;

	if (*ch1 > *ch2) {
		return 1;
	} else if (*ch1 == *ch2) {
		return 0;
	}
	return -1;
}

void cptblSort()
{
	qsort( (void*)map, TOTAL_TBL, sizeof(map[0]), mapSort );
}


//==============================================================
// VtgJISUCSϊ
//--------------------------------------------------------------

//===== R[hϊ =====

int sjis2ucs(int *uni, int sjis)
{
	static int		cache = 0;
	int				temp = sjis & 0xFF, low, hi, pos;
	unsigned short	*p;

	if (cache) {
		temp = (temp << 8) | cache;
		cache = 0x00;
	} else if (IS_SJIS_ZENKAKU(temp)) {
		cache = temp;
		*uni = -1;
		return 1;
	}

	if (temp <= 0x7F) {											// asciȉꍇ̓V[gJbg
		*uni = temp;
		return 0;
	}

	low = 0;
	hi = TOTAL_TBL;
	while (hi - low > 1) {
		pos = low + ((hi - low) >>1);
		p = map[pos];
		if (*p == temp) {										// ړI̕𔭌
			*uni = *(p + 1);
			return 0;
		} else if (*p > temp) {
			hi = pos;
		} else {
			low = pos;
		}
	}
	*uni = 0;													// ړI͌̕Ȃ
	return -1;
}

//===== ϊ =====

void psp2chSJIS2UCS(unsigned short *dst, const unsigned char *src, unsigned int num)
{
	int		sjis, len = 0;

	while (*src && (unsigned int)len < num) {
		sjis2ucs( &sjis, *src++ );
		if (sjis == -1) {
			continue;
		}
		*dst++ = (unsigned short)sjis;
		len++;
	}
	*dst = '\0';
}


//==============================================================
// UCSVtgJISϊ
//--------------------------------------------------------------

//===== R[hϊ =====

int ucs2sjis(int *sjis, int uni)
{
	int	i;
	for (i = 0; i < TOTAL_TBL; i++) {
		if (map[i][1] == uni) break;
	}
	if (i == TOTAL_TBL) {
		*sjis = 0;
		return -1;
	}
	*sjis = map[i][0];
	return 0;
}

//===== ϊ =====

void psp2chUCS2SJIS(unsigned char *dst, const unsigned short *src, unsigned int num)
{
	unsigned int	sjis, len = 0;

	while (*src) {
		if (ucs2sjis((int*)&sjis, (int)(*src++))) {
			sjis = 0xA181;										// ϊłȂ''
		}
        if (sjis < 0x100)
        {
            *dst++ = sjis;
            len++;
        }
        else
        {
            if (len >= num - 2)
            {
                break;
            }
            *dst++ = sjis & 0xFF;
            *dst++ = (sjis>>8) & 0xFF;
            len += 2;
        }
        if (len >= num - 1)
        {
            break;
        }
    }
    *dst = '\0';
}

/**********************
̃GR[hϊ
Shift JISEUC-JP
**********************/
void psp2chSjisToEuc(char* dst, char* src)
{
    unsigned char hi, lo;

    while (*src)
    {
        // pJi
        if ((unsigned char)*src > 0xA0 && (unsigned char)*src < 0xE0)
        {
            *dst++ = 0x8E;
            *dst++ = *src++;
        }
        // 2oCg
        else if ((unsigned char)*src > 0x7F)
        {
            hi = *src++;
            lo = *src++;
            hi -= (hi <= 0x9f) ? 0x71 : 0xb1;
            hi <<= 1;
            hi++;
            if (lo > 0x7f)
                lo--;
            if (lo >= 0x9e) {
                lo -= 0x7d;
                hi++;
            }
            else {
                lo -= 0x1f;
            }
            *dst++ = hi | 0x80;
            *dst++ = lo | 0x80;
        }
        // ASCII
        else
        {
            *dst++ = *src++;
        }
    }
    *dst = '\0';
}

/**********************
̃GR[hϊ
EUC-JPShift JIS
**********************/
void psp2chEucToSjis(char* dst, char* src)
{
    unsigned char hi, lo;

    while (*src)
    {
    	// X212(݂Ȃ"H"֕ϊ
    	if ((unsigned char)*src == 0x8F)
    	{
    		src += 3;
    		*dst++ = 0x81;
    		*dst++ = 0x48;
    	}
        // pJi
        else if ((unsigned char)*src == 0x8E)
        {
            src++;
            *dst++ = *src++;
        }
        // 2oCg
        else if ((unsigned char)*src > 0x7F)
        {
            hi = *src++ & 0x7F;
            lo = *src++ & 0x7F;
            hi -= 0x21;
            if (hi & 1)
            {
                lo += 0x7E;
            }
            else
            {
                lo += 0x1F;
                if (lo>= 0x7F)
                {
                    lo++;
                }
            }
            hi >>= 1;
            hi += (hi <= 0x1E) ? 0x81 : 0xC1;
            *dst++ = hi;
            *dst++ = lo;
        }
        // ASCII
        else
        {
            *dst++ = *src++;
        }
    }
    *dst = '\0';
}

//==============================================================
// UCSUTF8֕R[hϊ
//--------------------------------------------------------------

//===== R[hϊ =====

int ucs2utf8(unsigned char *dst, unsigned short src)
{
	int		len;

	if (src <= 0x007F) {
		*dst++ = src;
		len = 1;
	} else if (src <= 0x07FF) {
		*dst++ = ((src >> 6) & 0x1F) | 0xC0;
		*dst++ = (src & 0x3F) | 0x80;
		len = 2;
	} else {
		*dst++ = (src >> 12) | 0xE0;
		*dst++ = ((src >> 6) & 0x3F) | 0x80;
		*dst++ = (src & 0x3F) | 0x80;
		len = 3;
	}

	return len;
}

//===== ϊiTCYj =====

int UCS2UTF8Len(unsigned char *dst, int size, const unsigned short *src)
{
	unsigned char	utf8[3];
	int				i, len, step;

	len = 0;
	while (*src && len<size-1) {
		step = ucs2utf8( utf8, *src );
		if (size>=0 && len+step>=size) break;					// obt@TCY𒴂̂
		len += step;
		for (i = 0; i < step; i++) {
			*dst++ = utf8[i];
		}
		src++;
	}
	*dst = '\0';
	return len;
}

//===== ϊ =====

int psp2chUCS2UTF8(unsigned char *dst, const unsigned short *src)
{
	return UCS2UTF8Len( dst, -1, src );
}

/*********************
srcURLGR[hdstɊi[@
dst̗eʂɒ(src3{OK)
*********************/
void psp2chUrlEncode(char* dst, char* src)
{
    char buf[4];

    while (*src)
    {
        if ((*src >= 0x30 && *src <= 0x39) || // '0'-'9'
            (*src >= 0x41 && *src <= 0x59) || // 'A'-'Z'
            (*src >= 0x61 && *src <= 0x79) || // 'a'-'z'
            *src == '-' || *src == '.' || *src == '_')
        {
            *dst++ = *src++;
        }
        else if (*src == ' ')
        {
            *dst++ = '+';
            src++;
        }
        else
        {
            sprintf(buf, "%02X", *(unsigned char*)src++);
            *dst++ = '%';
            *dst++ = buf[0];
            *dst++ = buf[1];
        }
    }
    *dst = '\0';
}

//==============================================================
// UTF8VtgJIS֕R[hϊ
//--------------------------------------------------------------

//===== ϊiTCYj =====

void UTF82SjisLen(unsigned char *dst, int size, unsigned char *str)
{
	int		sjis, uni, len;

	len = 0;
	while (*str) {
		if (size>=0 && len>=size-1) break;						// obt@TCY𒴂̂
		if (*str <= 0x7F) {										// 1oCgR[hiASCIIj
			*dst++ = *str++;
			len++;
		} else {
			if (*str <= 0xDFU) {								// 2oCgR[h
				uni = ((str[0] & 0x001fu) << 6) | (str[1] & 0x003fu);
				str += 2;
			} else if (*str <= 0xEFU) {							// 3oCgR[h
				uni = ((str[0] & 0x001fu) << 12) | ((str[1] & 0x003fu) << 6) | (str[2] & 0x003fu);
				str += 3;
			} else {
				str++;
				uni = -1;
			}
			if (ucs2sjis(&sjis, uni) == -1) {
				sjis = 0x81A1;									// ϊłȂ''
			}
			*dst++ = sjis & 0xFF;
			len++;
			if (size<=0 && len>=size-1) {						// obt@TCY𒴂Ȃ
				dst--;
				break;
			}
			*dst++ = (sjis>>8) & 0xFF;
			len++;
		}
	}
	*dst = '\0';
}

//===== ϊ =====

void psp2chUTF82Sjis(unsigned char *dst, unsigned char *str)
{
	UTF82SjisLen( dst, -1, str );								// TCYw肹ɕϊ
}

//==============================================================
// VtgJISUTF8֕R[hϊ
//--------------------------------------------------------------

//===== ϊiTCYj =====
// dstSizeF荞ރobt@̃TCYi-1:j
// strSizeFƂȂ镶̒i-1:'\0'܂Łj
// ߂lF荞񂾌̒
//------------------------------------

int Sjis2UTF8Len(unsigned char *dst, int dstSize, unsigned char *str, int strSize)
{
	unsigned char	utf8[3];
	int				i, len, strLen, ucs, step;

	len = 0;
	strLen = 0;
	while (str[strLen]) {
		if (strSize>=0 && strLen>=strSize) break;				// ϊ񂪏I
		sjis2ucs( &ucs, str[strLen++] );
		if (ucs == -1) {										// QoCgR[ĥPoCgڂ
			continue;
		}
		step = ucs2utf8( utf8, ucs );
		if (dstSize>=0 && len+step>=dstSize) break;				// obt@TCY𒴂̂
		len += step;
		for (i = 0; i < step; i++) {
			*dst++ = utf8[i];
		}
	}
	*dst = '\0';
	if (str[strLen] == '\0') strLen++;

	return strLen;
}

//===== ϊ =====

void psp2chSjis2UTF8(unsigned char *dst, unsigned char *str)
{
	Sjis2UTF8Len( dst, -1, str, -1 );
}

//==============================================================
// Base64ϊ
//--------------------------------------------------------------

void psp2chBase64(char *dst, char *str)
{
	char			*table = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	unsigned char	*p;
	int				pos, dat, bit;

	p = (unsigned char*)str;
	pos = 0;
	dat = 0;
	bit = 0;
	for (; *p; p++) {
		dat = dat << 8 | *p;
		for (bit += 8; bit >= 6; bit -= 6) {
			dst[pos++] = table[(dat >> (bit - 6)) & 0x3f];
		}
	}
	if (bit > 0) {
		dat <<= 6 - bit;
		dst[pos++] = table[dat & 0x3f];
	}
	for (; pos % 4;) {
		dst[pos++] = '=';
	}
	dst[pos] = '\0';
}

